@@ -25,14 +25,14 @@ use std::collections::HashMap;
 
 use getopts::Options;
 
-use users::*;
+use uzers::*;
 
 /// walk through user ACL
 fn do_list(ro: &mut RunOptions, vec_eo: &[EnvOptions], service: &str) {
     let name = if ro.target == ro.name || ro.target.is_empty() {
-        "You"
+        "You".to_string()
     } else {
-        &ro.target
+        ro.target.clone()
     };
 
     let can_do = can(vec_eo, ro);
@@ -82,7 +82,7 @@ fn do_list(ro: &mut RunOptions, vec_eo: &[EnvOptions], service: &str) {
 /// navigate to directory or exit 1
 fn do_dir_changes(ro: &RunOptions, service: &str) {
     if ro.directory.is_some() {
-        if let Err(x) = std::env::set_current_dir(&ro.directory.as_ref().unwrap()) {
+        if let Err(x) = std::env::set_current_dir(ro.directory.as_ref().unwrap()) {
             println!(
                 "[{}] cannot cd into {}: {}",
                 &service,
@@ -96,7 +96,7 @@ fn do_dir_changes(ro: &RunOptions, service: &str) {
 
 /// setup getopts for argument parsing and help output
 fn general_options(
-    mut ro: &mut RunOptions,
+    ro: &mut RunOptions,
     args: Vec<String>,
     service: &str,
     vec_eo: &mut Vec<EnvOptions>,
@@ -143,10 +143,25 @@ fn general_options(
         ) as i32);
     }
 
+    let root_uid = nix::unistd::Uid::from_raw(0);
+    let root_gid = nix::unistd::Gid::from_raw(0);
+    if nix::unistd::getuid() != root_uid {
+        if !set_privs("root", root_uid, root_gid) {
+            std::process::exit(1);
+        }
+
+        if !drop_privs(ro) {
+            std::process::exit(1);
+        }
+    }
+
     if matches.opt_present("a") {
         let mut vec = vec![];
 
         for s in matches.opt_str("a").unwrap().split(',') {
+            if s.trim() == "" {
+                continue;
+            }
             vec.push(s.to_string());
         }
         ro.allow_env_list = Some(vec);
@@ -170,6 +185,28 @@ fn general_options(
     }
 }
 
+fn exit_if_command_not_found(ro: &RunOptions, service: &str) {
+    if let Some(k) = ro.located_bin.get(&ro.new_args[0]) {
+        if k.is_none() {
+            println!("[{service}] command not found");
+            std::process::exit(1);
+        }
+    }
+}
+
+fn is_command_cd(ro: &RunOptions, service: &str) {
+    if ro.cloned_args.is_none() && &ro.new_args[0] == "cd" {
+        println!("[{service}] {} is a shell feature.", &ro.new_args[0]);
+        if ro.new_args.len() > 1 {
+            println!(
+                "Try either changing to {} first or using {} -d {} instead.",
+                &ro.new_args[1], service, &ro.new_args[1]
+            );
+        }
+        std::process::exit(1);
+    }
+}
+
 /// main entry point
 fn main() {
     let args: Vec<String> = std::env::args().collect();
@@ -184,13 +221,14 @@ fn main() {
 
     let root_uid = nix::unistd::Uid::from_raw(0);
     let root_gid = nix::unistd::Gid::from_raw(0);
+    if nix::unistd::getuid() == root_uid {
+        if !set_privs("root", root_uid, root_gid) {
+            std::process::exit(1);
+        }
 
-    if !set_privs("root", root_uid, root_gid) {
-        std::process::exit(1);
-    }
-
-    if !drop_privs(&ro) {
-        std::process::exit(1);
+        if !drop_privs(&ro) {
+            std::process::exit(1);
+        }
     }
 
     general_options(&mut ro, args, &service, &mut vec_eo);
@@ -220,6 +258,8 @@ fn main() {
         std::process::exit(1);
     }
 
+    ro.command = replace_new_args(ro.new_args.clone());
+
     if ro.acl_type == Acltype::List {
         if ro.target.is_empty() {
             ro.target = ro.name.to_string();
@@ -232,37 +272,20 @@ fn main() {
         ro.target = "root".to_string();
     }
 
-    ro.command = replace_new_args(ro.new_args.clone());
-
-    match search_path(&ro.new_args[0]) {
-        None => {
-            println!("[{}]: command not found", service);
-            if &ro.new_args[0] == "cd" {
-                println!("{} is a shell feature.", &ro.new_args[0]);
-                if ro.new_args.len() > 1 {
-                    println!(
-                        "Try either changing to {} first or using {} -d {} instead.",
-                        &ro.new_args[1], service, &ro.new_args[1]
-                    );
-                }
-            }
-
-            std::process::exit(1);
-        }
-        Some(x) => {
-            ro.new_args[0] = x;
-            ro.command = replace_new_args(ro.new_args.clone());
-        }
-    }
-
-    let entry = can(&vec_eo, &ro);
+    let entry = can(&vec_eo, &mut ro);
     ro.env_options = Some(entry.clone());
 
     if entry.syslog.is_some() {
         ro.syslog = entry.syslog.unwrap();
     }
+
     if !entry.permit() {
         log_action(&service, "deny", &ro, &ro.original_command.join(" "));
+
+        is_command_cd(&ro, &service);
+
+        exit_if_command_not_found(&ro, &service);
+
         print_may_not(&ro);
         std::process::exit(1);
     }
@@ -273,6 +296,7 @@ fn main() {
         std::process::exit(1);
     }
 
+    // password required?
     if !challenge_password(&ro, &entry, &service) {
         log_action(&service, "deny", &ro, &ro.original_command.join(" "));
         std::process::exit(1);
@@ -282,6 +306,7 @@ fn main() {
         std::process::exit(1);
     }
 
+    // target user
     let lookup_name = get_user_by_name(&ro.target);
     if lookup_name.is_none() {
         println!("Could not lookup {}", &ro.target);
@@ -298,6 +323,7 @@ fn main() {
         std::process::exit(1);
     }
 
+    // change to target dir
     do_dir_changes(&ro, &service);
 
     if !drop_privs(&ro) {
@@ -318,11 +344,13 @@ fn main() {
 
     nix::sys::stat::umask(ro.old_umask.unwrap());
 
-    if ro.new_args.len() > 1 {
-        Command::new(&ro.new_args[0])
-            .args(ro.new_args.clone().split_off(1))
+    if ro.cloned_args.as_ref().unwrap().len() > 1 {
+        Command::new(&ro.cloned_args.as_ref().unwrap()[0])
+            .args(ro.cloned_args.as_ref().unwrap().clone().split_off(1))
             .exec();
     } else {
-        Command::new(&ro.new_args[0]).exec();
+        Command::new(&ro.cloned_args.as_ref().unwrap()[0]).exec();
     }
+    println!("Error executing");
+    std::process::exit(1);
 }
