@@ -33,11 +33,11 @@ use std::io::prelude::*;
 use std::io::BufReader;
 use std::os::unix::io::AsRawFd;
 use std::time::SystemTime;
-use users::os::unix::UserExt;
-use users::*;
+use uzers::os::unix::UserExt;
+use uzers::*;
 
 use getopts::{Matches, Options};
-use nix::unistd::{alarm, gethostname, initgroups, setegid, seteuid, setgid, setuid};
+use nix::unistd::{alarm, gethostname, setegid, seteuid, setgid, setuid};
 use pam::Authenticator;
 
 use rand::distributions::Alphanumeric;
@@ -55,7 +55,7 @@ pub enum ReasonType {
     Text(String),
 }
 
-#[derive(Clone)]
+#[derive(Clone, Debug)]
 pub struct EnvOptions {
     pub name: Option<String>,
     pub exact_name: Option<String>,
@@ -87,6 +87,8 @@ pub struct EnvOptions {
     pub env_permit: Option<String>,
     pub env_assign: Option<HashMap<String, String>>,
     pub timeout: Option<u32>,
+    pub search_path: Option<String>,
+    pub token_timeout: Option<u64>,
 }
 
 impl EnvOptions {
@@ -122,6 +124,8 @@ impl EnvOptions {
             env_permit: None,
             env_assign: None,
             timeout: None,
+            search_path: None,
+            token_timeout: None,
         }
     }
     fn new_deny() -> EnvOptions {
@@ -154,7 +158,7 @@ impl Default for EnvOptions {
     }
 }
 
-#[derive(Clone)]
+#[derive(Clone, Debug)]
 pub struct RunOptions {
     pub name: String,
     pub original_uid: nix::unistd::Uid,
@@ -178,6 +182,9 @@ pub struct RunOptions {
     pub old_envs: Option<HashMap<String, String>>,
     pub allow_env_list: Option<Vec<String>>,
     pub env_options: Option<EnvOptions>,
+    pub cloned_args: Option<Vec<String>>,
+    pub located_bin: HashMap<String, Option<String>>,
+    pub resume: Option<bool>,
 }
 
 impl RunOptions {
@@ -205,6 +212,9 @@ impl RunOptions {
             old_envs: None,
             allow_env_list: None,
             env_options: None,
+            cloned_args: None,
+            located_bin: HashMap::new(),
+            resume: None,
         }
     }
 }
@@ -226,10 +236,10 @@ impl pam::Converse for PamConvo {
         CString::new(self.login.clone()).map_err(|_| ())
     }
     fn prompt_blind(&mut self, _msg: &CStr) -> Result<CString, ()> {
-        match rpassword::read_password_from_tty(Some(&format!(
+        match rpassword::prompt_password(format!(
             "[{}] password for {}: ",
             self.service, self.login
-        ))) {
+        )) {
             Ok(password) => self.passwd = Some(password),
             Err(_) => {
                 println!("Cannot read from terminal");
@@ -454,12 +464,10 @@ pub fn common_opt_arguments(
         std::process::exit(0);
     }
 
-    let mut buf = [0u8; 64];
-    ro.hostname = gethostname(&mut buf)
+    ro.hostname = gethostname()
         .expect("Failed getting hostname")
-        .to_str()
-        .expect("Hostname wasn't valid UTF-8")
-        .to_string();
+        .into_string()
+        .expect("Hostname wasn't valid UTF-8");
 }
 
 /// read an ini file and traverse includes
@@ -566,11 +574,10 @@ pub fn read_ini(
                         }
                         collect.sort();
                         for file in collect {
-                            let incf = file;
-                            if !can_dir_include(&incf) {
+                            if !can_dir_include(&file) {
                                 continue;
                             }
-                            if read_ini_config_file(&incf, vec_eo, ro, fail_error, bytes, ini_list)
+                            if read_ini_config_file(&file, vec_eo, ro, fail_error, bytes, ini_list)
                             {
                                 println!("Could not include file");
                                 return true;
@@ -658,14 +665,16 @@ pub fn read_ini(
                 opt.notbefore = Some(
                     parse_date_from_str(value, "%Y%m%d")
                         .unwrap()
-                        .and_hms(0, 0, 0),
+                        .and_hms_opt(0, 0, 0)
+                        .unwrap(),
                 )
             }
             "notafter" if value.len() == 8 => {
                 opt.notafter = Some(
                     parse_date_from_str(value, "%Y%m%d")
                         .unwrap()
-                        .and_hms(23, 59, 59),
+                        .and_hms_opt(23, 59, 59)
+                        .unwrap(),
                 )
             }
             "notbefore" if value.len() == 14 => {
@@ -743,7 +752,17 @@ pub fn read_ini(
                     opt.timeout = Some(timeout.unwrap());
                 }
             }
-
+            "search_path" => {
+                opt.search_path = Some(value.to_string());
+            }
+            "token_timeout" => {
+                let token_timeout: Result<u64, core::num::ParseIntError> = value.parse();
+                if fail_error && token_timeout.is_err() {
+                    faulty = true;
+                } else {
+                    opt.token_timeout = Some(token_timeout.unwrap());
+                }
+            }
             &_ => {
                 println!("Error parsing {}:{}", config_path, line_number);
                 faulty = true;
@@ -771,7 +790,7 @@ pub fn read_ini_config_file(
     let path = Path::new(config_path);
     let display = path.display();
 
-    let file = match File::open(&path) {
+    let file = match File::open(path) {
         Err(why) => {
             println!("Could not open {}: {}", display, why);
             return true;
@@ -960,6 +979,7 @@ pub fn rule_match(item: &EnvOptions, ro: &RunOptions, line: Option<i32>) -> bool
     if item.exact_rule.is_some() {
         let exact_rule = item.exact_rule.as_ref().unwrap();
         if exact_rule == &ro.command {
+            // println!("{}: exact rule match: {} == {}", item.section, exact_rule, ro.command);
             return true;
         }
         // println!("{}: exact rule mismatch: {} != {}", item.section, exact_rule, ro.command);
@@ -986,6 +1006,7 @@ pub fn rule_match(item: &EnvOptions, ro: &RunOptions, line: Option<i32>) -> bool
             // opt = item.clone();
             return true;
         }
+        // println!("{}: item rule is not match", &item.rule.as_ref().unwrap());
         return false;
     }
     false
@@ -1000,7 +1021,7 @@ pub fn directory_check_ok(item: &EnvOptions, ro: &RunOptions, line: Option<i32>)
 
         let exact_dir = item.exact_dir.as_ref().unwrap();
 
-        if (&ro.directory.as_ref()).is_some() && exact_dir != ro.directory.as_ref().unwrap() {
+        if (ro.directory.as_ref()).is_some() && exact_dir != ro.directory.as_ref().unwrap() {
             return false;
         }
         return true;
@@ -1025,7 +1046,7 @@ pub fn directory_check_ok(item: &EnvOptions, ro: &RunOptions, line: Option<i32>)
             }
         };
 
-        if (&ro.directory.as_ref()).is_some() && !dir_re.is_match(ro.directory.as_ref().unwrap()) {
+        if (ro.directory.as_ref()).is_some() && !dir_re.is_match(ro.directory.as_ref().unwrap()) {
             // && ro.directory != "." {
             return false;
         }
@@ -1184,7 +1205,7 @@ pub fn group_matches(item: &EnvOptions, ro: &RunOptions, line: Option<i32>) -> b
     false
 }
 
-pub fn matching(item: &EnvOptions, ro: &RunOptions, line_error: Option<i32>) -> bool {
+pub fn matching(item: &EnvOptions, ro: &mut RunOptions, line_error: Option<i32>) -> bool {
     if !permitted_dates_ok(item, ro, line_error) {
         // println!("Didn't match permitted dates");
         return false;
@@ -1230,6 +1251,34 @@ pub fn matching(item: &EnvOptions, ro: &RunOptions, line_error: Option<i32>) ->
         return true;
     }
 
+    // cloned_args and command should be reset each loop
+    // search_path could expose privilege paths that may appear
+    // in error messaging
+
+    // shouldn't matter if setuid running user happened first, but even
+    // so better than to be sorry later
+    if ro.cloned_args.is_some() {
+        ro.cloned_args = None;
+    }
+    ro.command = replace_new_args(ro.new_args.clone());
+
+    if item.acl_type == Acltype::Run {
+        match search_path(ro, item) {
+            None => {
+                return false;
+            }
+            Some(x) => {
+                ro.cloned_args = Some(ro.new_args.clone());
+                ro.cloned_args.as_mut().unwrap()[0] = x;
+                ro.command = replace_new_args(ro.cloned_args.as_ref().unwrap().clone());
+            }
+        }
+    }
+    if item.acl_type == Acltype::Edit {
+        let edit_file = vec![ro.new_args[0].clone()];
+        ro.command = replace_new_args(edit_file);
+    }
+
     rule_match(item, ro, line_error)
 }
 
@@ -1286,11 +1335,21 @@ pub fn merge_default(default: &EnvOptions, item: &EnvOptions) -> EnvOptions {
         merged.permit = default.permit;
     }
 
+    if default.search_path.is_some() && item.search_path.is_none() {
+        // println!("merging search_path");
+        merged.search_path = default.search_path.clone();
+    }
+
+    if default.token_timeout.is_some() && item.token_timeout.is_none() {
+        // println!("merging token_timeout");
+        merged.token_timeout = default.token_timeout;
+    }
+
     merged
 }
 
 /// search the EnvOptions list for matching RunOptions and return the match
-pub fn can(vec_eo: &[EnvOptions], ro: &RunOptions) -> EnvOptions {
+pub fn can(vec_eo: &[EnvOptions], ro: &mut RunOptions) -> EnvOptions {
     let mut opt = EnvOptions::new_deny();
     let mut default = EnvOptions::new();
 
@@ -1425,7 +1484,7 @@ pub fn challenge_password(ro: &RunOptions, entry: &EnvOptions, service: &str) ->
             std::process::exit(1);
         }
 
-        if valid_token(&ro.name) {
+        if valid_token(&ro.name, entry) {
             update_token(&ro.name);
             return true;
         }
@@ -1450,13 +1509,13 @@ pub fn challenge_password(ro: &RunOptions, entry: &EnvOptions, service: &str) ->
 
                 let tty = std::fs::File::open("/dev/tty");
                 if tty.is_ok() {
-                    let term_res = nix::sys::termios::tcgetattr(tty.as_ref().unwrap().as_raw_fd());
+                    let term_res = nix::sys::termios::tcgetattr(tty.as_ref().unwrap());
                     if let Ok(mut term) = term_res {
                         term.local_flags
                             .set(nix::sys::termios::LocalFlags::ECHO, true);
 
                         let res = nix::sys::termios::tcsetattr(
-                            tty.as_ref().unwrap().as_raw_fd(),
+                            tty.as_ref().unwrap(),
                             nix::sys::termios::SetArg::TCSANOW,
                             &term,
                         );
@@ -1656,24 +1715,73 @@ pub fn produce_list(vec_eo: &[EnvOptions], ro: &RunOptions) -> Vec<String> {
     str_list
 }
 
+/// return result from search cache lookup
+pub fn search_path_cache(ro: &RunOptions, binary: &str) -> Option<String> {
+    match ro.located_bin.get(binary) {
+        Some(k) => {
+            if k.is_none() {
+                // println!("{} returning None (cached lookup)", item.section);
+                return None;
+            }
+            // println!("{} returning Some({})", item.section, k.as_ref().unwrap().to_string());
+            return Some(k.as_ref().unwrap().to_string());
+        }
+        None => {
+            // println!("{} returning None (not a cached lookup)", item.section);
+            None
+        }
+    }
+}
+
 /// if binary is not an absolute/relative path, look for it in usual places
-pub fn search_path(binary: &str) -> Option<String> {
-    let p = Path::new(binary);
+pub fn search_path(ro: &mut RunOptions, item: &EnvOptions) -> Option<String> {
+    let binary = &ro.new_args[0];
+    let p = Path::new(&binary);
+    // println!("Searching for {binary}");
+
     if binary.starts_with('/') || binary.starts_with("./") {
+        let lookup = search_path_cache(ro, binary);
+        if lookup.is_some() {
+            return lookup;
+        }
+
         if !p.exists() {
+            ro.located_bin.insert(binary.to_string(), None);
             return None;
         } else {
+            ro.located_bin
+                .insert(binary.to_string(), Some(binary.to_string()));
             return Some(binary.to_string());
         }
     }
 
-    for dir in "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin".split(':') {
+    let dirs = if item.search_path.is_some() {
+        item.search_path.as_ref().unwrap()
+    } else {
+        "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin"
+    };
+
+    for dir in dirs.split(':') {
+        if dir.trim() == "" {
+            continue;
+        }
+        let dir = dir.trim_end_matches('/');
         let path_name = format!("{}/{}", &dir, &binary);
+
+        if let Some(lookup) = search_path_cache(ro, binary) {
+            return Some(lookup);
+        }
+
         let p = Path::new(&path_name);
 
         if !p.exists() {
+            ro.located_bin.insert(binary.to_string(), None);
             continue;
         }
+
+        // println!("inserting {binary} = {}", path_name.clone());
+        ro.located_bin
+            .insert(binary.to_string(), Some(path_name.clone()));
         return Some(path_name);
     }
 
@@ -1788,10 +1896,25 @@ pub fn bad_priv_msg() {
 
 /// set privs of usr to target_uid and target_gid. return false if fails
 pub fn set_privs(user: &str, target_uid: nix::unistd::Uid, target_gid: nix::unistd::Gid) -> bool {
-    let user = CString::new(user).unwrap();
-    if initgroups(&user, target_gid).is_err() {
-        bad_priv_msg();
-        return false;
+    let name = CString::new(user.as_bytes()).unwrap();
+
+    unsafe {
+        if libc::setgroups(0, std::ptr::null()) != 0 {
+            bad_priv_msg();
+            return false;
+        }
+
+        #[cfg(not(target_os = "macos"))]
+        if libc::initgroups(name.as_ptr(), target_gid.as_raw()) != 0 {
+            bad_priv_msg();
+            return false;
+        }
+
+        #[cfg(target_os = "macos")]
+        if libc::initgroups(name.as_ptr(), target_gid.as_raw() as i32) != 0 {
+            bad_priv_msg();
+            return false;
+        }
     }
 
     if setgid(target_gid).is_err() {
@@ -1940,7 +2063,7 @@ pub fn token_path(user: &str) -> Option<String> {
 }
 
 pub fn create_token_dir() -> bool {
-    if !Path::new(&token_dir()).is_dir() && fs::create_dir_all(&token_dir()).is_err() {
+    if !Path::new(&token_dir()).is_dir() && fs::create_dir_all(token_dir()).is_err() {
         println!("Could not create token directory");
         return false;
     }
@@ -1967,7 +2090,7 @@ pub fn boot_secs() -> libc::timespec {
 /// does the user have a valid token
 /// return false if time stamp is in the future
 /// return true if token was set within 600 seconds of wall and boot time
-pub fn valid_token(user: &str) -> bool {
+pub fn valid_token(user: &str, eo: &EnvOptions) -> bool {
     if !create_token_dir() {
         return false;
     }
@@ -1977,7 +2100,7 @@ pub fn valid_token(user: &str) -> bool {
         return false;
     }
 
-    let secs = 600;
+    let secs = eo.token_timeout.unwrap_or(600);
 
     let token_path = token_path.unwrap();
     match fs::metadata(token_path) {
@@ -2056,7 +2179,7 @@ pub fn update_token(user: &str) {
         return;
     }
 
-    if std::fs::rename(&token_path_tmp.as_str(), token_path).is_err() {
+    if std::fs::rename(token_path_tmp.as_str(), token_path).is_err() {
         println!("Could not update token");
     }
 }
@@ -2098,7 +2221,7 @@ pub fn group_hash(groups: Vec<Group>) -> HashMap<String, u32> {
 pub fn replace_new_args(new_args: Vec<String>) -> String {
     let mut args = vec![];
     for arg in &new_args {
-        args.push(arg.replace('\\', "\\\\").replace(' ', "\\ "));
+        args.push(arg.replace('\\', r"\\").replace(' ', r"\ "));
     }
 
     args.join(" ")
@@ -2118,7 +2241,7 @@ pub fn prng_alpha_num_string(n: usize) -> String {
         .collect()
 }
 
-pub fn runopt_target_gid(ro: &RunOptions, lookup_name: &users::User) -> nix::unistd::Gid {
+pub fn runopt_target_gid(ro: &RunOptions, lookup_name: &uzers::User) -> nix::unistd::Gid {
     if ro.target_group.is_some() {
         match nix::unistd::Group::from_name(ro.target_group.as_ref().unwrap()) {
             Ok(x) => match x {
